"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Minimist = require("minimist");
var git_1 = require("./git");
var local_storage_1 = require("./local-storage");
var step_1 = require("./step");
var submodule_1 = require("./submodule");
/**
 The rebase module is responsible for performing tasks done by the editor using an
 interactive rebase.
 */
function init() {
    if (require.main !== module) {
        return;
    }
    var argv = Minimist(process.argv.slice(2), {
        string: ['_'],
    });
    var method = argv._[0];
    var arg1 = argv._[1];
    switch (method) {
        case 'reword':
            return rewordRecentStep(arg1);
        case 'super-pick':
            return superPickStep(arg1);
        case 'rebranch-super':
            return rebranchSuperSteps();
    }
}
init();
// Responsible for editing the recent commit's message. It will also sort the step's
// number if needed
function rewordRecentStep(message) {
    // Calculate next step based on the current commit's message
    var commitMessage = git_1.Git.recentCommit(['--format=%s']);
    var stepDescriptor = step_1.Step.descriptor(commitMessage);
    var nextStep = getNextStep(stepDescriptor);
    // Open the editor by default
    var args = ['commit', '--amend', '--allow-empty'];
    // If message provided skip editor
    if (message) {
        args.push('-m', message);
    }
    // Specified step is gonna be used for when forming the commit message
    local_storage_1.localStorage.setItem('HOOK_STEP', nextStep);
    // This will be used later on to update the manuals
    if (stepDescriptor && step_1.Step.ensureStepMap()) {
        step_1.Step.updateStepMap('reset', { oldStep: stepDescriptor.number, newStep: nextStep });
    }
    // commit, let git hooks do the rest
    git_1.Git.print(args);
}
exports.rewordRecentStep = rewordRecentStep;
// The super-picker is responsible for cherry-picking super steps and
// rename their manual files
function superPickStep(hash) {
    var message = git_1.Git.recentCommit(hash, ['--format=%s']);
    var oldStep = step_1.Step.descriptor(message).number;
    var newStep = step_1.Step.nextSuper();
    // Fetch patch data
    var diff = newStep - oldStep;
    var stepFilePattern = /step(\d+)\.(md|tmpl)/g;
    var patch = git_1.Git(['format-patch', '-1', hash, '--stdout']);
    // Replace references for old manual files with new manual files
    var fixedPatch = patch.replace(stepFilePattern, function (file, step, extension) {
        step = Number(step) + diff;
        return "step" + step + "." + extension;
    });
    var submoduleCwd = local_storage_1.localStorage.getItem('SUBMODULE_CWD');
    var stepMap = step_1.Step.getStepMap(submoduleCwd, true);
    if (stepMap) {
        // The submodule cwd was given from another process of tortilla
        var submodule_2 = submodule_1.Submodule.getLocalName(submoduleCwd);
        var diffStepPattern = /\{\{\s*diffStep\s+"?(\d+\.\d+)"?.*\}\}/g;
        // Replace indexes presented in diffStep() template helpers
        fixedPatch = fixedPatch.replace(diffStepPattern, function (helper, oldStepId) {
            var helperSubmodule = helper.match(/module\s?=\s?"?([^\s"]+)"?/);
            helperSubmodule = helperSubmodule ? helperSubmodule[1] : '';
            // Don't replace anything if submodules don't match
            if (helperSubmodule !== submodule_2) {
                return helper;
            }
            // In case step has been removed in the process, replace it with a meaningless placeholder
            var newStepId = stepMap[oldStepId] || 'XX.XX';
            return helper.replace(/(diffStep\s+"?)\d+\.\d+/, "$1" + newStepId);
        });
    }
    // Apply patch
    git_1.Git(['am'], {
        input: fixedPatch,
    });
}
exports.superPickStep = superPickStep;
// Updates the branches referencing all super steps
function rebranchSuperSteps() {
    var rootBranch = git_1.Git.activeBranchName();
    // Delete root
    try {
        git_1.Git(['branch', '-D', rootBranch + "-root"]);
    }
    catch (e) {
        // Ignore
    }
    // Delete steps
    git_1.Git(['branch']).split('\n').filter(function (branch) {
        return branch.match(new RegExp(rootBranch + "-step\\d+"));
    })
        .forEach(function (branch) {
        git_1.Git(['branch', '-D', branch.trim()]);
    });
    // Branch root
    git_1.Git(['branch', rootBranch + "-root", git_1.Git.rootHash()]);
    // Branch steps
    git_1.Git(['log', '--format=%H %s', '--grep=^Step [0-9]\\+:'])
        .split('\n')
        .filter(Boolean)
        .map(function (log) {
        var message = log.split(' ');
        var hash = message.shift();
        message = message.join(' ');
        return {
            number: step_1.Step.descriptor(message).number,
            hash: hash,
        };
    })
        .forEach(function (step) {
        git_1.Git(['branch', rootBranch + "-step" + step.number, step.hash]);
    });
}
exports.rebranchSuperSteps = rebranchSuperSteps;
// Calculate the next step dynamically based on its super flag
function getNextStep(stepDescriptor) {
    if (!stepDescriptor) {
        return '';
    }
    var isSubStep = !!stepDescriptor.number.split('.')[1];
    return isSubStep ? step_1.Step.next(1) : step_1.Step.nextSuper(1);
}
//# sourceMappingURL=rebase.js.map